@@ -14,7 +14,7 @@
 # You should have received a copy of the GNU Affero General Public License
 # along with this program.  If not, see <https://www.gnu.org/licenses/>.
 from typing import (Dict, Tuple, Optional, List, Deque, Set, Any, Union, AsyncGenerator,
-                    NamedTuple, TYPE_CHECKING, cast)
+                    Awaitable, NamedTuple, TYPE_CHECKING, cast)
 from collections import deque
 from datetime import datetime
 from os import path
@@ -33,6 +33,7 @@ from mautrix.types import (EventID, MessageEventContent, RoomID, EventType, Mess
                            ContentURI, EncryptedFile)
 from mautrix.errors import MatrixError, MForbidden
 from mautrix.util.simple_lock import SimpleLock
+from mautrix.util.network_retry import call_with_net_retry
 
 from .db import Portal as DBPortal, Message as DBMessage, Reaction as DBReaction
 from .config import Config
@@ -60,6 +61,7 @@ class Portal(DBPortal, BasePortal):
     config: Config
     matrix: 'm.MatrixHandler'
     az: AppService
+    private_chat_portal_meta: bool
 
     _main_intent: Optional[IntentAPI]
     _create_room_lock: asyncio.Lock
@@ -105,6 +107,7 @@ class Portal(DBPortal, BasePortal):
         cls.az = bridge.az
         cls.loop = bridge.loop
         cls.bridge = bridge
+        cls.private_chat_portal_meta = cls.config["bridge.private_chat_portal_meta"]
         NotificationDisabler.puppet_cls = p.Puppet
         NotificationDisabler.config_enabled = cls.config["bridge.backfill.disable_notifications"]
 
@@ -308,8 +311,8 @@ class Portal(DBPortal, BasePortal):
             upload_mime_type = "application/octet-stream"
             upload_file_name = None
 
-        mxc = await intent.upload_media(data, mime_type=upload_mime_type,
-                                        filename=upload_file_name)
+        mxc = await call_with_net_retry(intent.upload_media, data, mime_type=upload_mime_type,
+                                        filename=upload_file_name, _action="upload media")
 
         if decryption_info:
             decryption_info.url = mxc
@@ -373,7 +376,10 @@ class Portal(DBPortal, BasePortal):
             puppet = await p.Puppet.get_by_twid(self.other_user)
             if not self._main_intent:
                 self._main_intent = puppet.default_mxid_intent
-            changed = await self._update_name(puppet.name)
+            if self.encrypted or self.private_chat_portal_meta:
+                changed = await self._update_name(puppet.name)
+            else:
+                changed = False
         else:
             changed = await self._update_name(conv.name)
         if changed:
@@ -381,6 +387,16 @@ class Portal(DBPortal, BasePortal):
             await self.update()
         await self._update_participants(conv.participants)
 
+    async def update_name(self, name: str) -> None:
+        if not self.encrypted and not self.private_chat_portal_meta:
+            return
+
+        changed = await self._update_name(name)
+
+        if changed:
+            await self.update_bridge_info()
+            await self.update()
+
     async def _update_name(self, name: str) -> bool:
         if self.name != name:
             self.name = name
@@ -593,7 +609,7 @@ class Portal(DBPortal, BasePortal):
             })
             if self.is_direct:
                 invites.append(self.az.bot_mxid)
-        if self.encrypted or not self.is_direct:
+        if self.encrypted or self.private_chat_portal_meta or not self.is_direct:
             name = self.name
         if self.config["appservice.community_id"]:
             initial_state.append({
@@ -622,15 +638,16 @@ class Portal(DBPortal, BasePortal):
             self.by_mxid[self.mxid] = self
             if not self.is_direct:
                 await self._update_participants(info.participants)
-            else:
-                puppet = await p.Puppet.get_by_custom_mxid(source.mxid)
-                if puppet:
-                    try:
-                        await puppet.intent.join_room_by_id(self.mxid)
+
+            puppet = await p.Puppet.get_by_custom_mxid(source.mxid)
+            if puppet:
+                try:
+                    await puppet.intent.join_room_by_id(self.mxid)
+                    if self.is_direct:
                         await source.update_direct_chats({self.main_intent.mxid: [self.mxid]})
-                    except MatrixError:
-                        self.log.debug("Failed to join custom puppet into newly created portal",
-                                       exc_info=True)
+                except MatrixError:
+                    self.log.debug("Failed to join custom puppet into newly created portal",
+                                   exc_info=True)
 
             if not info.trusted:
                 msg = "This is a message request. Replying here will accept the request."
@@ -673,9 +690,16 @@ class Portal(DBPortal, BasePortal):
         await self.update()
 
     @classmethod
-    async def all_with_room(cls) -> AsyncGenerator['Portal', None]:
-        portals = await super().all_with_room()
-        portal: cls
+    def all_with_room(cls) -> AsyncGenerator['Portal', None]:
+        return cls._db_to_portals(super().all_with_room())
+
+    @classmethod
+    def find_private_chats_with(cls, other_user: int) -> AsyncGenerator['Portal', None]:
+        return cls._db_to_portals(super().find_private_chats_with(other_user))
+
+    @classmethod
+    async def _db_to_portals(cls, query: Awaitable[List['Portal']]) -> AsyncGenerator['Portal', None]:
+        portals = await query
         for index, portal in enumerate(portals):
             try:
                 yield cls.by_twid[(portal.twid, portal.receiver)]
